@@ -1,21 +1,20 @@
-from abc import abstractmethod
-from operator import itemgetter
-from typing import List, Dict, Union, Callable
-import re
-
 import torch, flair
 import logging
+import re
+
+from abc import abstractmethod, ABC
 
 from collections import Counter
 from collections import defaultdict
 
-from segtok.segmenter import split_single
-from segtok.tokenizer import split_contractions
-from segtok.tokenizer import word_tokenizer
-from torch.utils.data import Dataset, random_split
+from deprecated import deprecated
+from flair.file_utils import Tqdm
+from operator import itemgetter
+
+from torch.utils.data import Dataset
 from torch.utils.data.dataset import ConcatDataset, Subset
 
-from flair.file_utils import Tqdm
+from typing import List, Dict, Union, Callable, Optional
 
 log = logging.getLogger("flair")
 
@@ -113,20 +112,17 @@ class Dictionary:
     @classmethod
     def load(cls, name: str):
         from flair.file_utils import cached_path
-
+        hu_path: str = "https://flair.informatik.hu-berlin.de/resources/characters"
         if name == "chars" or name == "common-chars":
-            base_path = "https://s3.eu-central-1.amazonaws.com/alan-nlp/resources/models/common_characters"
-            char_dict = cached_path(base_path, cache_dir="datasets")
+            char_dict = cached_path(f"{hu_path}/common_characters", cache_dir="datasets")
             return Dictionary.load_from_file(char_dict)
 
         if name == "chars-large" or name == "common-chars-large":
-            base_path = "https://s3.eu-central-1.amazonaws.com/alan-nlp/resources/models/common_characters_large"
-            char_dict = cached_path(base_path, cache_dir="datasets")
+            char_dict = cached_path(f"{hu_path}/common_characters_large", cache_dir="datasets")
             return Dictionary.load_from_file(char_dict)
 
         if name == "chars-xl" or name == "common-chars-xl":
-            base_path = "https://s3.eu-central-1.amazonaws.com/alan-nlp/resources/models/common_characters_xl"
-            char_dict = cached_path(base_path, cache_dir="datasets")
+            char_dict = cached_path(f"{hu_path}/common_characters_xl", cache_dir="datasets")
             return Dictionary.load_from_file(char_dict)
 
         return Dictionary.load_from_file(name)
@@ -219,6 +215,10 @@ class DataPoint:
 
         return self
 
+    def remove_labels(self, label_type: str):
+        if label_type in self.annotation_layers.keys():
+            del self.annotation_layers[label_type]
+
     def get_labels(self, label_type: str = None):
         if label_type is None:
             return self.labels
@@ -338,17 +338,18 @@ class Token(DataPoint):
                 if name in self._embeddings.keys():
                     del self._embeddings[name]
 
-    def get_each_embedding(self) -> torch.tensor:
+    def get_each_embedding(self, embedding_names: Optional[List[str]] = None) -> torch.tensor:
         embeddings = []
         for embed in sorted(self._embeddings.keys()):
+            if embedding_names and embed not in embedding_names: continue
             embed = self._embeddings[embed].to(flair.device)
             if (flair.embedding_storage_mode == "cpu") and embed.device != flair.device:
                 embed = embed.to(flair.device)
             embeddings.append(embed)
         return embeddings
 
-    def get_embedding(self) -> torch.tensor:
-        embeddings = self.get_each_embedding()
+    def get_embedding(self, names: Optional[List[str]] = None) -> torch.tensor:
+        embeddings = self.get_each_embedding(names)
 
         if embeddings:
             return torch.cat(embeddings, dim=0)
@@ -442,6 +443,15 @@ class Span(DataPoint):
             else '<span ({}): "{}">'.format(ids, self.text)
         )
 
+    def __getitem__(self, idx: int) -> Token:
+        return self.tokens[idx]
+
+    def __iter__(self):
+        return iter(self.tokens)
+
+    def __len__(self) -> int:
+        return len(self.tokens)
+
     @property
     def tag(self):
         return self.labels[0].value
@@ -451,26 +461,85 @@ class Span(DataPoint):
         return self.labels[0].score
 
 
+class Tokenizer(ABC):
+    r"""An abstract class representing a :class:`Tokenizer`.
+
+    Tokenizers are used to represent algorithms and models to split plain text into
+    individual tokens / words. All subclasses should overwrite :meth:`tokenize`, which
+    splits the given plain text into tokens. Moreover, subclasses may overwrite
+    :meth:`name`, returning a unique identifier representing the tokenizer's
+    configuration.
+    """
+
+    @abstractmethod
+    def tokenize(self, text: str) -> List[Token]:
+        raise NotImplementedError()
+
+    @property
+    def name(self) -> str:
+        return self.__class__.__name__
+
+
+@deprecated(version="0.5", reason="Use 'flair.tokenization.SpaceTokenizer' instead.")
+def space_tokenizer(text: str) -> List[Token]:
+    # We don't want to create a SpaceTokenizer object each time this function is called,
+    # so delegate the call directly to the static run_tokenize method
+    from flair.tokenization import SpaceTokenizer
+    return SpaceTokenizer.run_tokenize(text)
+
+
+@deprecated(version="0.5", reason="Use 'flair.tokenization.SegtokTokenizer' instead.")
+def segtok_tokenizer(text: str) -> List[Token]:
+    # We don't want to create a SegtokTokenizer object each time this function is called,
+    # so delegate the call directly to the static run_tokenize method
+    from flair.tokenization import SegtokTokenizer
+    return SegtokTokenizer.run_tokenize(text)
+
+
+@deprecated(version="0.5", reason="Use 'flair.tokenization.SpacyTokenizer' instead.")
+def build_spacy_tokenizer(model) -> Callable[[str], List[Token]]:
+    from flair.tokenization import SpacyTokenizer
+    spacy_tokenizer = SpacyTokenizer(model)
+
+    def tokenizer(text: str) -> List[Token]:
+        return spacy_tokenizer.tokenize(text)
+
+    return tokenizer
+
+
+@deprecated(version="0.5", reason="Use 'flair.tokenization.JapaneseTokenizer' instead.")
+def build_japanese_tokenizer(tokenizer: str = "MeCab"):
+    from flair.tokenization import JapaneseTokenizer
+    japanese_tokenizer = JapaneseTokenizer(tokenizer)
+
+    def tokenizer(text: str) -> List[Token]:
+        return japanese_tokenizer.tokenize(text)
+
+    return tokenizer
+
+
 class Sentence(DataPoint):
     """
-       A Sentence is a list of Tokens and is used to represent a sentence or text fragment.
+       A Sentence is a list of tokens and is used to represent a sentence or text fragment.
     """
 
     def __init__(
         self,
         text: str = None,
-        use_tokenizer: Union[bool, Callable[[str], List[Token]]] = False,
+        use_tokenizer: Union[bool, Tokenizer] = True,
         language_code: str = None,
+        start_position: int = None
     ):
         """
         Class to hold all meta related to a text (tokens, predictions, language code, ...)
         :param text: original string
-        :param use_tokenizer: a custom tokenizer (default is space based tokenizer,
-        more advanced options are segtok_tokenizer to use segtok or build_spacy_tokenizer to use Spacy library
-        if available). Check the code of space_tokenizer to implement your own (if you need it).
-        If instead of providing a function, this parameter is just set to True, segtok will be used.
-        :param labels:
-        :param language_code:
+        :param use_tokenizer: a custom tokenizer (default is :class:`SpaceTokenizer`)
+            more advanced options are :class:`SegTokTokenizer` to use segtok or :class:`SpacyTokenizer`
+            to use Spacy library if available). Check the implementations of abstract class Tokenizer or
+            implement your own subclass (if you need it). If instead of providing a Tokenizer, this parameter
+            is just set to True (deprecated), :class:`SegtokTokenizer` will be used.
+        :param language_code: Language of the sentence
+        :param start_position: Start char offset of the sentence in the superordinate document
         """
         super().__init__()
 
@@ -480,19 +549,32 @@ class Sentence(DataPoint):
 
         self.language_code: str = language_code
 
-        tokenizer = use_tokenizer
-        if type(use_tokenizer) == bool:
-            tokenizer = segtok_tokenizer if use_tokenizer else space_tokenizer
+        self.start_pos = start_position
+        self.end_pos = (
+            start_position + len(text) if start_position is not None else None
+        )
+
+        if isinstance(use_tokenizer, Tokenizer):
+            tokenizer = use_tokenizer
+        elif hasattr(use_tokenizer, "__call__"):
+            from flair.tokenization import TokenizerWrapper
+            tokenizer = TokenizerWrapper(use_tokenizer)
+        elif type(use_tokenizer) == bool:
+            from flair.tokenization import SegtokTokenizer, SpaceTokenizer
+            tokenizer = SegtokTokenizer() if use_tokenizer else SpaceTokenizer()
+        else:
+            raise AssertionError("Unexpected type of parameter 'use_tokenizer'. " +
+                                 "Parameter should be bool, Callable[[str], List[Token]] (deprecated), Tokenizer")
 
         # if text is passed, instantiate sentence with tokens (words)
         if text is not None:
             text = self._restore_windows_1252_characters(text)
-            [self.add_token(token) for token in tokenizer(text)]
+            [self.add_token(token) for token in tokenizer.tokenize(text)]
 
         # log a warning if the dataset is empty
         if text == "":
             log.warning(
-                "ACHTUNG: An empty Sentence was created! Are there empty strings in your dataset?"
+                "Warning: An empty Sentence was created! Are there empty strings in your dataset?"
             )
 
         self.tokenized = None
@@ -529,9 +611,7 @@ class Sentence(DataPoint):
             label_names.append(label.value)
         return label_names
 
-    def get_spans(self, label_type: str, min_score=-1) -> List[Span]:
-
-        spans: List[Span] = []
+    def _add_spans_internal(self, spans: List[Span], label_type: str, min_score):
 
         current_span = []
 
@@ -603,6 +683,24 @@ class Sentence(DataPoint):
 
         return spans
 
+    def get_spans(self, label_type: Optional[str] = None, min_score=-1) -> List[Span]:
+
+        spans: List[Span] = []
+
+        # if label type is explicitly specified, get spans for this label type
+        if label_type:
+            return self._add_spans_internal(spans, label_type, min_score)
+
+        # else determine all label types in sentence and get all spans
+        label_types = []
+        for token in self:
+            for annotation in token.annotation_layers.keys():
+                if annotation not in label_types: label_types.append(annotation)
+
+        for label_type in label_types:
+            self._add_spans_internal(spans, label_type, min_score)
+        return spans
+
     @property
     def embedding(self):
         return self.get_embedding()
@@ -615,9 +713,10 @@ class Sentence(DataPoint):
             vector = vector.to(device)
         self._embeddings[name] = vector
 
-    def get_embedding(self) -> torch.tensor:
+    def get_embedding(self, names: Optional[List[str]] = None) -> torch.tensor:
         embeddings = []
         for embed in sorted(self._embeddings.keys()):
+            if names and embed not in names: continue
             embedding = self._embeddings[embed]
             embeddings.append(embedding)
 
@@ -669,6 +768,8 @@ class Sentence(DataPoint):
 
                 if token.get_labels(label_type)[0].value == "O":
                     continue
+                if token.get_labels(label_type)[0].value == "_":
+                    continue
 
                 tags.append(token.get_labels(label_type)[0].value)
             all_tags = "<" + "/".join(tags) + ">"
@@ -771,6 +872,9 @@ class Sentence(DataPoint):
     def __iter__(self):
         return iter(self.tokens)
 
+    def __len__(self) -> int:
+        return len(self.tokens)
+
     def __repr__(self):
         tagged_string = self.to_tagged_string()
         tokenized_string = self.to_tokenized_string()
@@ -810,9 +914,6 @@ class Sentence(DataPoint):
 
         return f'Sentence: "{tokenized_string}"   [− Tokens: {len(self)}{token_labels}{sentence_labels}]'
 
-    def __len__(self) -> int:
-        return len(self.tokens)
-
     def get_language_code(self) -> str:
         if self.language_code is None:
             import langdetect
@@ -914,17 +1015,17 @@ class Corpus:
         if test is None:
             train_length = len(train)
             test_size: int = round(train_length / 10)
-            splits = random_split(train, [train_length - test_size, test_size])
-            train = splits[0]
-            test = splits[1]
+            splits = randomly_split_into_two_datasets(train, test_size)
+            test = splits[0]
+            train = splits[1]
 
         # sample dev data if none is provided
         if dev is None:
             train_length = len(train)
             dev_size: int = round(train_length / 10)
-            splits = random_split(train, [train_length - dev_size, dev_size])
-            train = splits[0]
-            dev = splits[1]
+            splits = randomly_split_into_two_datasets(train, dev_size)
+            dev = splits[0]
+            train = splits[1]
 
         # set train dev and test data
         self._train: FlairDataset = train
@@ -963,6 +1064,37 @@ class Corpus:
         self._dev = Corpus._filter_empty_sentences(self._dev)
         log.info(self)
 
+    def filter_long_sentences(self, max_charlength: int):
+        log.info("Filtering long sentences")
+        self._train = Corpus._filter_long_sentences(self._train, max_charlength)
+        self._test = Corpus._filter_long_sentences(self._test, max_charlength)
+        self._dev = Corpus._filter_long_sentences(self._dev, max_charlength)
+        log.info(self)
+
+    @staticmethod
+    def _filter_long_sentences(dataset, max_charlength: int) -> Dataset:
+
+        # find out empty sentence indices
+        empty_sentence_indices = []
+        non_empty_sentence_indices = []
+        index = 0
+
+        from flair.datasets import DataLoader
+
+        for batch in DataLoader(dataset):
+            for sentence in batch:
+                if len(sentence.to_plain_string()) > max_charlength:
+                    empty_sentence_indices.append(index)
+                else:
+                    non_empty_sentence_indices.append(index)
+                index += 1
+
+        # create subset of non-empty sentence indices
+        subset = Subset(dataset, non_empty_sentence_indices)
+
+        return subset
+
+
     @staticmethod
     def _filter_empty_sentences(dataset) -> Dataset:
 
@@ -1027,8 +1159,8 @@ class Corpus:
     def _downsample_to_proportion(dataset: Dataset, proportion: float):
 
         sampled_size: int = round(len(dataset) * proportion)
-        splits = random_split(dataset, [len(dataset) - sampled_size, sampled_size])
-        return splits[1]
+        splits = randomly_split_into_two_datasets(dataset, sampled_size)
+        return splits[0]
 
     def obtain_statistics(
             self, label_type: str = None, pretty_print: bool = True
@@ -1180,7 +1312,9 @@ class MultiCorpus(Corpus):
         )
 
     def __str__(self):
-        return "\n".join([str(corpus) for corpus in self.corpora])
+        output = f"MultiCorpus: {len(self.train)} train + {len(self.dev)} dev + {len(self.test)} test sentences\n - "
+        output += "\n - ".join([f'{type(corpus).__name__} {str(corpus)}' for corpus in self.corpora])
+        return output
 
 
 def iob2(tags):
@@ -1227,183 +1361,15 @@ def iob_iobes(tags):
             raise Exception("Invalid IOB format!")
     return new_tags
 
+def randomly_split_into_two_datasets(dataset, length_of_first):
 
-def space_tokenizer(text: str) -> List[Token]:
-    """
-    Tokenizer based on space character only.
-    """
-    tokens: List[Token] = []
-    word = ""
-    index = -1
-    for index, char in enumerate(text):
-        if char == " ":
-            if len(word) > 0:
-                start_position = index - len(word)
-                tokens.append(
-                    Token(
-                        text=word, start_position=start_position, whitespace_after=True
-                    )
-                )
+    import random
+    indices = [i for i in range(len(dataset))]
+    random.shuffle(indices)
 
-            word = ""
-        else:
-            word += char
-    # increment for last token in sentence if not followed by whitespace
-    index += 1
-    if len(word) > 0:
-        start_position = index - len(word)
-        tokens.append(
-            Token(text=word, start_position=start_position, whitespace_after=False)
-        )
-    return tokens
+    first_dataset = indices[:length_of_first]
+    second_dataset = indices[length_of_first:]
+    first_dataset.sort()
+    second_dataset.sort()
 
-
-def build_japanese_tokenizer(tokenizer: str = "MeCab"):
-    if tokenizer.lower() != "mecab":
-        raise NotImplementedError("Currently, MeCab is only supported.")
-
-    try:
-        import konoha
-    except ModuleNotFoundError:
-        log.warning("-" * 100)
-        log.warning('ATTENTION! The library "konoha" is not installed!')
-        log.warning(
-            'To use Japanese tokenizer, please first install with the following steps:'
-        )
-        log.warning(
-            '- Install mecab with "sudo apt install mecab libmecab-dev mecab-ipadic"'
-        )
-        log.warning('- Install konoha with "pip install konoha[mecab]"')
-        log.warning("-" * 100)
-        pass
-
-    sentence_tokenizer = konoha.SentenceTokenizer()
-    word_tokenizer = konoha.WordTokenizer(tokenizer)
-
-    def tokenizer(text: str) -> List[Token]:
-        """
-        Tokenizer using konoha, a third party library which supports
-        multiple Japanese tokenizer such as MeCab, KyTea and SudachiPy.
-        """
-        tokens: List[Token] = []
-        words: List[str] = []
-
-        sentences = sentence_tokenizer.tokenize(text)
-        for sentence in sentences:
-            konoha_tokens = word_tokenizer.tokenize(sentence)
-            words.extend(list(map(str, konoha_tokens)))
-
-        # determine offsets for whitespace_after field
-        index = text.index
-        current_offset = 0
-        previous_word_offset = -1
-        previous_token = None
-        for word in words:
-            try:
-                word_offset = index(word, current_offset)
-                start_position = word_offset
-            except:
-                word_offset = previous_word_offset + 1
-                start_position = (
-                    current_offset + 1 if current_offset > 0 else current_offset
-                )
-
-            token = Token(
-                text=word, start_position=start_position, whitespace_after=True
-            )
-            tokens.append(token)
-
-            if (previous_token is not None) and word_offset - 1 == previous_word_offset:
-                previous_token.whitespace_after = False
-
-            current_offset = word_offset + len(word)
-            previous_word_offset = current_offset - 1
-            previous_token = token
-
-        return tokens
-
-    return tokenizer
-
-
-def segtok_tokenizer(text: str) -> List[Token]:
-    """
-    Tokenizer using segtok, a third party library dedicated to rules-based Indo-European languages.
-    https://github.com/fnl/segtok
-    """
-    tokens: List[Token] = []
-
-    words: List[str] = []
-    sentences = split_single(text)
-    for sentence in sentences:
-        contractions = split_contractions(word_tokenizer(sentence))
-        words.extend(contractions)
-
-    # determine offsets for whitespace_after field
-    index = text.index
-    current_offset = 0
-    previous_word_offset = -1
-    previous_token = None
-    for word in words:
-        try:
-            word_offset = index(word, current_offset)
-            start_position = word_offset
-        except:
-            word_offset = previous_word_offset + 1
-            start_position = (
-                current_offset + 1 if current_offset > 0 else current_offset
-            )
-
-        if word:
-            token = Token(
-                text=word, start_position=start_position, whitespace_after=True
-            )
-            tokens.append(token)
-
-        if (previous_token is not None) and word_offset - 1 == previous_word_offset:
-            previous_token.whitespace_after = False
-
-        current_offset = word_offset + len(word)
-        previous_word_offset = current_offset - 1
-        previous_token = token
-
-    return tokens
-
-
-def build_spacy_tokenizer(model) -> Callable[[str], List[Token]]:
-    """
-    Wrap Spacy model to build a tokenizer for the Sentence class.
-    :param model a Spacy V2 model
-    :return a tokenizer function to provide to Sentence class constructor
-    """
-    try:
-        from spacy.language import Language
-        from spacy.tokens.doc import Doc
-        from spacy.tokens.token import Token as SpacyToken
-    except ImportError:
-        raise ImportError(
-            "Please install Spacy v2.0 or better before using the Spacy tokenizer, otherwise you can use segtok_tokenizer as advanced tokenizer."
-        )
-
-    model: Language = model
-
-    def tokenizer(text: str) -> List[Token]:
-        doc: Doc = model.make_doc(text)
-        previous_token = None
-        tokens: List[Token] = []
-        for word in doc:
-            word: SpacyToken = word
-            token = Token(
-                text=word.text, start_position=word.idx, whitespace_after=True
-            )
-            tokens.append(token)
-
-            if (previous_token is not None) and (
-                    token.start_pos - 1
-                    == previous_token.start_pos + len(previous_token.text)
-            ):
-                previous_token.whitespace_after = False
-
-            previous_token = token
-        return tokens
-
-    return tokenizer
+    return [Subset(dataset, first_dataset), Subset(dataset, second_dataset)]
