@@ -1,20 +1,21 @@
 import logging
 from pathlib import Path
-from typing import List, Union, Callable, Dict
+from typing import List, Union, Dict, Optional
 
 import torch
 import torch.nn as nn
-from torch.utils.data import DataLoader
+from torch.utils.data.dataset import Dataset
 from tqdm import tqdm
+import numpy as np
 
+import sklearn.metrics as metrics
 import flair.nn
 import flair.embeddings
-from flair.data import Dictionary, Sentence, Label, Token, space_tokenizer, DataPoint
-from flair.datasets import SentenceDataset, StringDataset
+from flair.data import Dictionary, Sentence, Label, DataPoint
+from flair.datasets import SentenceDataset, DataLoader
 from flair.file_utils import cached_path
 from flair.training_utils import (
     convert_labels_to_one_hot,
-    Metric,
     Result,
     store_embeddings,
 )
@@ -97,8 +98,10 @@ class TextClassifier(flair.nn.Model):
 
         self.document_embeddings.embed(sentences)
 
+        embedding_names = self.document_embeddings.get_names()
+
         text_embedding_list = [
-            sentence.embedding.unsqueeze(0) for sentence in sentences
+            sentence.get_embedding(embedding_names).unsqueeze(0) for sentence in sentences
         ]
         text_embedding_tensor = torch.cat(text_embedding_list, 0).to(flair.device)
 
@@ -111,6 +114,7 @@ class TextClassifier(flair.nn.Model):
             "state_dict": self.state_dict(),
             "document_embeddings": self.document_embeddings,
             "label_dictionary": self.label_dictionary,
+            "label_type": self.label_type,
             "multi_label": self.multi_label,
             "beta": self.beta,
             "weight_dict": self.weight_dict,
@@ -152,40 +156,36 @@ class TextClassifier(flair.nn.Model):
 
     def predict(
         self,
-        sentences: Union[List[Sentence], Sentence, List[str], str],
+        sentences: Union[List[Sentence], Sentence],
         mini_batch_size: int = 32,
-        embedding_storage_mode="none",
         multi_class_prob: bool = False,
         verbose: bool = False,
-        use_tokenizer: Union[bool, Callable[[str], List[Token]]] = space_tokenizer,
-    ) -> List[Sentence]:
+        label_name: Optional[str] = None,
+        return_loss = False,
+        embedding_storage_mode="none",
+    ):
         """
         Predicts the class labels for the given sentences. The labels are directly added to the sentences.
         :param sentences: list of sentences
         :param mini_batch_size: mini batch size to use
-        :param embedding_storage_mode: 'none' for the minimum memory footprint, 'cpu' to store embeddings in Ram,
-        'gpu' to store embeddings in GPU memory.
         :param multi_class_prob : return probability for all class for multiclass
         :param verbose: set to True to display a progress bar
-        :param use_tokenizer: a custom tokenizer when string are provided (default is space based tokenizer).
-        :return: the list of sentences containing the labels
+        :param return_loss: set to True to return loss
+        :param label_name: set this to change the name of the label type that is predicted
+        :param embedding_storage_mode: default is 'none' which is always best. Only set to 'cpu' or 'gpu' if
+        you wish to not only predict, but also keep the generated embeddings in CPU or GPU memory respectively.
+        'gpu' to store embeddings in GPU memory.
         """
-        predicted_label_type = self.label_type if self.label_type is not None else 'class'
+        if label_name == None:
+            label_name = self.label_type if self.label_type is not None else 'label'
 
         with torch.no_grad():
             if not sentences:
                 return sentences
 
-            if isinstance(sentences, DataPoint) or isinstance(sentences, str):
+            if isinstance(sentences, DataPoint):
                 sentences = [sentences]
 
-            if (flair.device.type == "cuda") and embedding_storage_mode == "cpu":
-                log.warning(
-                    "You are inferring on GPU with parameter 'embedding_storage_mode' set to 'cpu'."
-                    "This option will slow down your inference, usually 'none' (default value) "
-                    "is a better choice."
-                )
-
             # filter empty sentences
             if isinstance(sentences[0], Sentence):
                 sentences = [sentence for sentence in sentences if len(sentence) > 0]
@@ -195,85 +195,99 @@ class TextClassifier(flair.nn.Model):
             rev_order_len_index = sorted(
                 range(len(sentences)), key=lambda k: len(sentences[k]), reverse=True
             )
-            original_order_index = sorted(
-                range(len(rev_order_len_index)), key=lambda k: rev_order_len_index[k]
-            )
 
             reordered_sentences: List[Union[DataPoint, str]] = [
                 sentences[index] for index in rev_order_len_index
             ]
 
-            if isinstance(sentences[0], DataPoint):
-                # remove previous embeddings
-                store_embeddings(reordered_sentences, "none")
-                dataset = SentenceDataset(reordered_sentences)
-            else:
-                dataset = StringDataset(
-                    reordered_sentences, use_tokenizer=use_tokenizer
-                )
             dataloader = DataLoader(
-                dataset=dataset, batch_size=mini_batch_size, collate_fn=lambda x: x
+                dataset=SentenceDataset(reordered_sentences), batch_size=mini_batch_size
             )
-
             # progress bar for verbosity
             if verbose:
                 dataloader = tqdm(dataloader)
 
-            results: List[Sentence] = []
-            for i, batch in enumerate(dataloader):
+            overall_loss = 0
+            batch_no = 0
+            for batch in dataloader:
+
+                batch_no += 1
+
                 if verbose:
-                    dataloader.set_description(f"Inferencing on batch {i}")
-                results += batch
+                    dataloader.set_description(f"Inferencing on batch {batch_no}")
+
                 # stop if all sentences are empty
                 if not batch:
                     continue
 
                 scores = self.forward(batch)
+
+                if return_loss:
+                    overall_loss += self._calculate_loss(scores, batch)
+
                 predicted_labels = self._obtain_labels(
                     scores, predict_prob=multi_class_prob
                 )
 
                 for (sentence, labels) in zip(batch, predicted_labels):
                     for label in labels:
-                        sentence.add_label(predicted_label_type, label.value, label.score)
+                        if self.multi_label or multi_class_prob:
+                            sentence.add_label(label_name, label.value, label.score)
+                        else:
+                            sentence.set_label(label_name, label.value, label.score)
 
                 # clearing token embeddings to save memory
                 store_embeddings(batch, storage_mode=embedding_storage_mode)
 
-            results: List[Union[Sentence, str]] = [
-                results[index] for index in original_order_index
-            ]
-            assert len(sentences) == len(results)
-            return results
+            if return_loss:
+                return overall_loss / batch_no
 
     def evaluate(
         self,
-        data_loader: DataLoader,
-        out_path: Path = None,
+        sentences: Union[List[DataPoint], Dataset],
+        out_path: Union[str, Path] = None,
         embedding_storage_mode: str = "none",
+        mini_batch_size: int = 32,
+        num_workers: int = 8,
     ) -> (Result, float):
 
+        # read Dataset into data loader (if list of sentences passed, make Dataset first)
+        if not isinstance(sentences, Dataset):
+            sentences = SentenceDataset(sentences)
+        data_loader = DataLoader(sentences, batch_size=mini_batch_size, num_workers=num_workers)
+
+        # use scikit-learn to evaluate
+        y_true = []
+        y_pred = []
+
         with torch.no_grad():
             eval_loss = 0
 
-            metric = Metric("Evaluation", beta=self.beta)
-
             lines: List[str] = []
             batch_count: int = 0
             for batch in data_loader:
 
                 batch_count += 1
 
-                scores = self.forward(batch)
-                predictions = self._obtain_labels(scores)
-                loss = self._calculate_loss(scores, batch)
+                # remove previously predicted labels
+                [sentence.remove_labels('predicted') for sentence in batch]
+
+                # get the gold labels
+                true_values_for_batch = [sentence.get_labels(self.label_type) for sentence in batch]
+
+                # predict for batch
+                loss = self.predict(batch,
+                                    embedding_storage_mode=embedding_storage_mode,
+                                    mini_batch_size=mini_batch_size,
+                                    label_name='predicted',
+                                    return_loss=True)
 
                 eval_loss += loss
 
                 sentences_for_batch = [sent.to_plain_string() for sent in batch]
 
-                true_values_for_batch = [sentence.get_labels(self.label_type) for sentence in batch]
-                available_labels = self.label_dictionary.get_items()
+                # get the predicted labels
+                predictions = [sentence.get_labels('predicted') for sentence in batch]
 
                 for sentence, prediction, true_value in zip(
                     sentences_for_batch,
@@ -292,56 +306,70 @@ class TextClassifier(flair.nn.Model):
                     true_values_for_sentence = [label.value for label in true_values_for_sentence]
                     predictions_for_sentence = [label.value for label in predictions_for_sentence]
 
-                    for label in available_labels:
-                        if (
-                            label in predictions_for_sentence
-                            and label in true_values_for_sentence
-                        ):
-                            metric.add_tp(label)
-                        elif (
-                            label in predictions_for_sentence
-                            and label not in true_values_for_sentence
-                        ):
-                            metric.add_fp(label)
-                        elif (
-                            label not in predictions_for_sentence
-                            and label in true_values_for_sentence
-                        ):
-                            metric.add_fn(label)
-                        elif (
-                            label not in predictions_for_sentence
-                            and label not in true_values_for_sentence
-                        ):
-                            metric.add_tn(label)
+                    y_true_instance = np.zeros(len(self.label_dictionary), dtype=int)
+                    for i in range(len(self.label_dictionary)):
+                        if self.label_dictionary.get_item_for_index(i) in true_values_for_sentence:
+                            y_true_instance[i] = 1
+                    y_true.append(y_true_instance.tolist())
+
+                    y_pred_instance = np.zeros(len(self.label_dictionary), dtype=int)
+                    for i in range(len(self.label_dictionary)):
+                        if self.label_dictionary.get_item_for_index(i) in predictions_for_sentence:
+                            y_pred_instance[i] = 1
+                    y_pred.append(y_pred_instance.tolist())
 
                 store_embeddings(batch, embedding_storage_mode)
 
-            eval_loss /= batch_count
-
-            detailed_result = (
-                f"\nMICRO_AVG: acc {metric.micro_avg_accuracy()} - f1-score {metric.micro_avg_f_score()}"
-                f"\nMACRO_AVG: acc {metric.macro_avg_accuracy()} - f1-score {metric.macro_avg_f_score()}"
-            )
-            for class_name in metric.get_classes():
-                detailed_result += (
-                    f"\n{class_name:<10} tp: {metric.get_tp(class_name)} - fp: {metric.get_fp(class_name)} - "
-                    f"fn: {metric.get_fn(class_name)} - tn: {metric.get_tn(class_name)} - precision: "
-                    f"{metric.precision(class_name):.4f} - recall: {metric.recall(class_name):.4f} - "
-                    f"accuracy: {metric.accuracy(class_name):.4f} - f1-score: "
-                    f"{metric.f_score(class_name):.4f}"
-                )
-
-            result = Result(
-                main_score=metric.micro_avg_accuracy(),
-                log_line=f"{metric.precision()}\t{metric.recall()}\t{metric.micro_avg_f_score()}",
-                log_header="PRECISION\tRECALL\tF1",
-                detailed_results=detailed_result,
-            )
+            # remove predicted labels
+            for sentence in sentences:
+                sentence.annotation_layers['predicted'] = []
 
             if out_path is not None:
                 with open(out_path, "w", encoding="utf-8") as outfile:
                     outfile.write("".join(lines))
 
+            # make "classification report"
+            target_names = []
+            for i in range(len(self.label_dictionary)):
+                target_names.append(self.label_dictionary.get_item_for_index(i))
+            classification_report = metrics.classification_report(y_true, y_pred, digits=4,
+                                                                  target_names=target_names, zero_division=0)
+
+            # get scores
+            micro_f_score = round(metrics.fbeta_score(y_true, y_pred, beta=self.beta, average='micro', zero_division=0), 4)
+            accuracy_score = round(metrics.accuracy_score(y_true, y_pred), 4)
+            macro_f_score = round(metrics.fbeta_score(y_true, y_pred, beta=self.beta, average='macro', zero_division=0), 4)
+            precision_score = round(metrics.precision_score(y_true, y_pred, average='macro', zero_division=0), 4)
+            recall_score = round(metrics.recall_score(y_true, y_pred, average='macro', zero_division=0), 4)
+
+            detailed_result = (
+                    "\nResults:"
+                    f"\n- F-score (micro) {micro_f_score}"
+                    f"\n- F-score (macro) {macro_f_score}"
+                    f"\n- Accuracy {accuracy_score}"
+                    '\n\nBy class:\n' + classification_report
+            )
+
+            # line for log file
+            if not self.multi_label:
+                log_header = "ACCURACY"
+                log_line = f"\t{accuracy_score}"
+            else:
+                log_header = "PRECISION\tRECALL\tF1\tACCURACY"
+                log_line = f"{precision_score}\t" \
+                           f"{recall_score}\t" \
+                           f"{macro_f_score}\t" \
+                           f"{accuracy_score}"
+
+            result = Result(
+                main_score=micro_f_score,
+                log_line=log_line,
+                log_header=log_header,
+                detailed_results=detailed_result,
+            )
+
+            eval_loss /= batch_count
+
             return result, eval_loss
 
     @staticmethod
@@ -431,15 +459,10 @@ class TextClassifier(flair.nn.Model):
     def _fetch_model(model_name) -> str:
 
         model_map = {}
-        aws_resource_path = "https://s3.eu-central-1.amazonaws.com/alan-nlp/resources/models-v0.4"
         hu_path: str = "https://nlp.informatik.hu-berlin.de/resources/models"
 
         model_map["de-offensive-language"] = "/".join(
-            [
-                aws_resource_path,
-                "classy-offensive-de-rnn-cuda%3A0",
-                "germ-eval-2018-task-1-v0.4.pt",
-            ]
+            [hu_path, "de-offensive-language", "germ-eval-2018-task-1-v0.4.pt"]
         )
 
         # English sentiment models
@@ -452,6 +475,11 @@ class TextClassifier(flair.nn.Model):
         model_map["sentiment-fast"] = "/".join(
             [hu_path, "sentiment-curated-fasttext-rnn", "sentiment-en-mix-ft-rnn.pt"]
         )
+        
+        #Communicative Functions Model
+        model_map["communicative-functions"] = "/".join(
+            [hu_path, "comfunc", "communicative-functions-v0.5b.pt"]
+        )
 
         cache_dir = Path("models")
         if model_name in model_map:
